"""

Tools for handling environment variables etc.

"""
import os
from collections.abc import Callable
from datetime import date, datetime
from typing import Any, Dict

from fmtr.tools.datatype_tools import to_bool
from fmtr.tools.path_tools import Path
from fmtr.tools.tools import identity, EMPTY


class MissingEnvironmentVariable(KeyError):
    """

    Exception for when a required environment variable is missing.

    """


def get_dict() -> Dict[str, str]:
    """

    Return environment variables as a standard dictionary.

    """
    environment_dict = dict(os.environ)
    return environment_dict


def get(name: str, default: Any = EMPTY, converter: Callable = identity, convert_empty: bool = False) -> Any:
    """

    Return the specified environment variable, handling default substitution and simple type conversion.

    """
    value = os.getenv(name, default)

    if value is EMPTY:
        msg = f'Environment variable "{name}" is required but has not been set'
        raise MissingEnvironmentVariable(msg)

    if value is not None or convert_empty:
        value = converter(value)

    return value


def get_getter(converter: Callable) -> Callable:
    """

    Return an environment getter for the specified type.

    """

    def func(name: str, default: Any = EMPTY):
        """

        Environment getter that converts to the specified type

        """
        value = get(name, default=default, converter=converter)
        return value

    return func


get_int = get_getter(lambda n: int(float(n)))
get_float = get_getter(float)
get_bool = get_getter(to_bool)
get_date = get_getter(date.fromisoformat)
get_datetime = get_getter(datetime.fromisoformat)
get_path = get_getter(Path)



import pytest
from datetime import datetime, date

from fmtr.tools import environment_tools
from fmtr.tools.path_tools import Path
from fmtr.tools.tests import helpers
from fmtr.tools.tools import EMPTY

ENVIRONMENT_TEST_DATA = [
    ('BOOLEAN1', 'Y', environment_tools.get_bool, True),
    ('BOOLEAN2', 'f', environment_tools.get_bool, False),
    ('DATE', '2001-01-01', environment_tools.get_date, date(2001, 1, 1)),
    ('DATETIME', '2021-10-09T10:25:51.425296', environment_tools.get_datetime,
     datetime(2021, 10, 9, 10, 25, 51, 425296)),
    ('INT1', '101', environment_tools.get_int, 101),
    ('INT2', '4.0E+10', environment_tools.get_int, 4.0E+10),
    ('FLOAT1', '123.456', environment_tools.get_float, 123.456),
    ('PATH', '/usr/bin/python', environment_tools.get_path, Path('/usr/bin/python'))
]

ENVIRONMENT_DATA = {key: raw for key, raw, func, expected in ENVIRONMENT_TEST_DATA}


@helpers.parametrize(
    'key,expected',
    ENVIRONMENT_DATA.items()
)
def test_get_env(key, expected):
    """

    """
    with helpers.patch_environment(clear=True, **ENVIRONMENT_DATA):
        actual = environment_tools.get(key)
    assert actual == expected


@helpers.parametrize(
    'expected',
    [None, 101, EMPTY]
)
def test_get_env_missing_raises(expected):
    """

    """
    MISSING_KEY = '!!MISSING!!'
    with helpers.patch_environment(clear=True, **ENVIRONMENT_DATA):
        if expected is EMPTY:
            with pytest.raises(environment_tools.MissingEnvironmentVariable):
                environment_tools.get(MISSING_KEY)
        else:
            actual = environment_tools.get(MISSING_KEY, default=expected)
            assert actual == expected


@helpers.parametrize(
    'key,raw,func,expected',
    ENVIRONMENT_TEST_DATA
)
def test_get_env_type(key, raw, func, expected):
    """

    """
    with helpers.patch_environment(clear=True, **ENVIRONMENT_DATA):
        actual = func(key)
    assert actual == expected


def test_get_env_dict():
    """

    """
    expected = ENVIRONMENT_DATA
    with helpers.patch_environment(clear=True, **ENVIRONMENT_DATA):
        actual = environment_tools.get_dict()
    assert actual == expected



import platform
from functools import lru_cache


@lru_cache
def is_wsl() -> bool:
    """

    Determine whether the platform is WSL or native Linux.

    """
    uname_data = platform.uname()
    is_wsl = uname_data.system.lower() == 'linux' and 'microsoft' in uname_data.release.lower()
    return is_wsl



import sys
from os import getenv

import importlib
import logging
import subprocess

from fmtr.tools import netrc_tools


def run():
    FMTR_LOG_LEVEL = getenv('FMTR_LOG_LEVEL', 'INFO')
    logging.getLogger().setLevel(FMTR_LOG_LEVEL)

    MODULE_NAME = getenv('PACKAGE_NAME')
    if not MODULE_NAME:
        raise KeyError('No MODULE_NAME set.')

    PIP_INDEX_URL = getenv('PIP_INDEX_URL')
    if not PIP_INDEX_URL:
        raise KeyError('No PIP_INDEX_URL set.')

    PIP_USERNAME = getenv('PIP_USERNAME')
    if not PIP_USERNAME:
        raise KeyError('No PIP_USERNAME set.')

    PIP_PASSWORD = getenv('PIP_PASSWORD')
    if not PIP_PASSWORD:
        raise KeyError('No PIP_PASSWORD set.')

    with netrc_tools.get() as netrc_obj:
        netrc_obj[PIP_INDEX_URL] = {
            netrc_tools.LOGIN: PIP_USERNAME,
            netrc_tools.PASSWORD: PIP_PASSWORD
        }

    print(f'Starting {MODULE_NAME}...')

    subprocess.run(
        [sys.executable, '-m', 'pip', 'install', MODULE_NAME, '--no-input', '--index-url', f'https://{PIP_INDEX_URL}'],
        check=True
    )

    interface = importlib.import_module(f'{MODULE_NAME}.interface')
    interface.run()



from pathlib import Path
from setuptools import find_namespace_packages, setup, find_packages

import requirements

name_ns, name_base = 'fmtr', 'tools'
name = f'{name_ns}.{name_base}'
path_segment = f'{name_ns}/{name_base}'
path_base = Path(__file__).absolute().parent
path_ver = path_base / path_segment / 'version'
__version__ = path_ver.read_text().strip()

packages = find_packages(where=path_segment)
packages = [name] + [f'{name}.{nsp}' for nsp in packages]

setup(
    name=name,
    version=__version__,
    url=f'https://github.com/{name_ns}/{name}',
    license='Copyright © 2024 Frontmatter. All rights reserved.',
    author='Frontmatter',
    author_email='innovative.fowler@mask.pro.fmtr.dev',
    description='Collection of high-level tools to simplify everyday development tasks, with a focus on AI/ML',
    long_description=(path_base / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    packages=find_namespace_packages(),
    package_dir={'': '.'},
    package_data={
        name: [f'version'],
    },
    install_requires=requirements.INSTALL,
    extras_require=requirements.EXTRAS,
)



from collections import namedtuple

from string import Formatter
from typing import List

formatter = Formatter()

Segment = namedtuple('Segment', ['literal_text', 'field_name', 'format_spec', 'conversion'])


def parse_string(string: str) -> List[Segment]:
    """

    Return structured version of a string with formatting slots.

    """
    parsed = [Segment(*args) for args in formatter.parse(string)]
    return parsed


def is_format_string(string: str) -> bool:
    """

    Does the string contains string formatting slots (i.e. {})?

    """
    try:
        parsed = parse_string(string)
    except ValueError:
        return False
    if all(datum.field_name is None for datum in parsed):
        return False
    else:
        return True


def get_var_name(string: str) -> str:
    """

    Get the name of a variable from a (resolved) f-string `{a=}`

    """
    name, value = string.split('=', maxsplit=1)
    return name



from typing import Tuple


def combine_args_kwargs(args: dict=None, kwargs: dict=None) -> dict:
    """

    Combines arguments and keyword arguments into a single dictionary.

    """
    args = args or []
    kwargs = kwargs or {}
    args = {i: arg for i, arg in enumerate(args)}
    args.update(kwargs)
    if all(isinstance(key, int) for key in args.keys()):
        args = list(args.values())
    return args


def split_args_kwargs(args_kwargs: dict) -> Tuple[list, dict]:
    """

    Splits arguments and keyword arguments into a list and a dictionary.

    """
    if isinstance(args_kwargs, list):
        args, kwargs = args_kwargs, {}
    else:
        args = [arg for key, arg in args_kwargs.items() if isinstance(key, int)]
        kwargs = {key: arg for key, arg in args_kwargs.items() if not isinstance(key, int)}

    return args, kwargs


